from cavass._log import logger
from cavass.ops import get_slice_number


def get_slice_range(partial_region_file: str, entire_region_file: str):
    """
    Get the slice location range where the `portion file` is located in the `entire region file`.

    Args:
        partial_region_file (str):
        entire_region_file (str):

    Returns:

    """
    slice_number_1 = get_slice_number(partial_region_file)
    slice_number_2 = get_slice_number(entire_region_file)
    # inferior slice refers to the index of slice in input_file_2 where
    # the inferior slice (the first slice) of input_file_1 located in input_file_2.
    # While the superior slice indicates the slice index in input_file_2
    # where the superior slice (the last slice) of input_file_1 located in input_file_2.
    inferior_slice_idx = round((slice_number_1[5] - slice_number_2[5]) / slice_number_2[2])
    superior_slice_idx = inferior_slice_idx + int(slice_number_1[8])
    inferior_slice_idx += 1
    if (slice_number_1[0] != slice_number_2[0]) or (slice_number_1[1] != slice_number_2[1]) or (
            slice_number_1[2] != slice_number_2[2]) or (slice_number_1[3] != slice_number_2[3]) or (
            slice_number_1[4] != slice_number_2[4]) or (slice_number_1[6] != slice_number_2[6]) or (
            slice_number_1[7] != slice_number_2[7]):
        logger.warning(
            f"Input files do not match.\nInput file 1 is {partial_region_file}.\nInput file 2 is {entire_region_file}.")
        unmatched = [partial_region_file, entire_region_file]
    else:
        unmatched = []
    return inferior_slice_idx, superior_slice_idx, unmatched



import os
import shutil

from cavass._io import ensure_output_file_dir_existence
from cavass.constants import CAVASS_START_INDEX
from cavass.ops import matched_reslice
from cavass.slice_range import get_slice_range


def match(unmatched_file: str,
          file_to_match: str,
          output_file: str):
    """
    Match body region.
    Args:
        unmatched_file (str):
        file_to_match (str):
        output_file (str):

    Returns:

    """
    inferior_slice_idx, superior_slice_idx, unmatched_files = get_slice_range(unmatched_file, file_to_match)
    inferior_slice_idx -= CAVASS_START_INDEX
    superior_slice_idx -= CAVASS_START_INDEX

    made_output_dir, output_dir = ensure_output_file_dir_existence(output_file)
    file_type = os.path.splitext(unmatched_file[0])[1][1:]
    interpolation_method = "nearest" if file_type == "BIM" else "linear"
    try:
        matched_reslice(unmatched_file, file_to_match, output_file, interpolation_method=interpolation_method)
    except Exception as e:
        if made_output_dir and os.path.isdir(output_dir):
            shutil.rmtree(output_dir)
        if os.path.exists(output_file):
            os.remove(output_file)
        raise e
    return unmatched_files



import nibabel as nib
from nibabel.orientations import io_orientation, axcodes2ornt, ornt_transform, inv_ornt_aff
import numpy as np


def reorient_nifti(input_nifti_image, target_orientation):
    """
    Reorient a NIfTI image.

    Parameters:
        input_nifti_image (): NIfTI image to be reoriented.
        target_orientation (str): Target orientation, e.g., 'ARI', 'RAI', 'LPI'.
    """
    data, affine, header = input_nifti_image.get_fdata(), input_nifti_image.affine, input_nifti_image.header

    # Current and target orientation
    current_ornt = io_orientation(affine)
    target_ornt = axcodes2ornt(tuple(target_orientation))

    # Check if reorientation is needed
    if np.array_equal(current_ornt, target_ornt):
        print(f"Image already in {target_orientation} orientation.")
        return input_nifti_image

    # Compute transform
    transform = ornt_transform(current_ornt, target_ornt)

    # Apply flips
    for i, (src, dir) in enumerate(transform):
        if dir == -1:
            data = np.flip(data, axis=int(src))

    # Transpose axes
    data = np.transpose(data, transform[:, 0].astype(int))

    # Update affine
    new_affine = inv_ornt_aff(transform, input_nifti_image.shape)
    return nib.Nifti1Image(data, new_affine, header=header)



import os
import shutil
from typing import Optional
from uuid import uuid1

import numpy as np

from cavass._io import ensure_output_file_dir_existence, save_nifti
from cavass.dicom import Modality
from cavass.ops import execute_cmd, get_voxel_spacing, read_cavass_file, copy_pose, get_image_resolution
from cavass.utils import one_hot

from cavass.nifti2dicom import nifti2dicom


def dicom2cavass(input_dicom_dir, output_file, offset=0, copy_pose_file=None):
    """
    Note that if the output file path is too long, this command may be failed.

    Args:
        input_dicom_dir (str):
        output_file (str):
        offset (int, optional, default=0):
        copy_pose_file (str, optional, default=None): if `copy_pose_file` is given, copy pose of this
        file to the output file.

    """

    if not os.path.exists(input_dicom_dir):
        raise ValueError(f"Input DICOM series {input_dicom_dir} not found.")
    if copy_pose_file is not None:
        if not os.path.isfile(copy_pose_file):
            raise FileNotFoundError(f"Copy pose file {copy_pose_file} not found.")

    made_output_dir, output_dir = ensure_output_file_dir_existence(output_file)

    input_dicom_dir = input_dicom_dir.replace(" ", "\ ")
    output_file = output_file.replace(" ", "\ ")
    try:
        if copy_pose_file is None:
            r = execute_cmd(f"from_dicom {input_dicom_dir}/* {output_file} +{offset}")
        else:
            split = os.path.splitext(output_file)
            root = split[0]
            extension = split[1]
            output_tmp_file = root + "_TMP" + extension
            r = execute_cmd(f"from_dicom {input_dicom_dir}/* {output_tmp_file} +{offset}")
            copy_pose(output_tmp_file, copy_pose_file, output_file)

    except Exception as e:
        if made_output_dir and os.path.isdir(output_dir):
            shutil.rmtree(output_dir)

        if copy_pose_file is not None and os.path.exists(output_tmp_file):
            os.remove(output_tmp_file)

        if os.path.exists(output_file):
            os.remove(output_file)
        raise e
    return r


def cavass2dicom(input_file, output_dicom_file, overwrite, start_slice: Optional[int] = None, end_slice: Optional[int] = None):
    """
    Convert CAVASS file (IM0 and BIM files) to DICOM series.
    Args:
        input_file (str):
        output_dicom_file (str): Output DICOM series filename without extension.
        overwrite (bool): Remove the output directory if it already exists if True. Otherwise, the convertion will be failed if the output directory already exists.
        start_slice (int, optional, default=None): The start slice number for conversion. If `None` (default), conversion starts from the first slice.
        end_slice (int, optional, default=None): The end slice number for conversion. If `None` (default), conversion continues to the final slice.

    Returns:

    """
    if not os.path.isfile(input_file):
        raise FileNotFoundError(f"Input file {input_file} not found.")

    if start_slice is None:
        start_slice = 0
    else:
        start_slice = start_slice - 1

    total_slice_number = get_image_resolution(input_file)[2]
    if end_slice is None:
        end_slice = total_slice_number - 1
    else:
        end_slice = end_slice - 1

    if start_slice < 0 or start_slice >= total_slice_number:
        raise ValueError(f"Start slice {start_slice} is out of bounds: [0, {total_slice_number - 1}].")

    if end_slice < 0 or end_slice >= total_slice_number:
        raise ValueError(f"End slice {end_slice} is out of bounds: [0, {total_slice_number - 1}].")

    if start_slice > end_slice:
        raise ValueError(f"Start slice {start_slice} must be less than end slice {end_slice}.")

    made_output_dir, output_dir = ensure_output_file_dir_existence(output_dicom_file)

    try:
        execute_cmd(f"mipg2dicom \"{input_file}\" \"{output_dicom_file}\" 0 {start_slice} {end_slice}")
    except Exception as e:
        if made_output_dir and os.path.isdir(output_dir):
            shutil.rmtree(output_dir)
        raise e


def nifti2cavass(input_nifti_file, output_file, modality, offset=0, copy_pose_file=None):
    """
    Convert NIfTI image to cavass image.

    Args:
        input_nifti_file (str):
        output_file (str):
        modality (Modality):
        offset (int, optional, default=0):
        copy_pose_file (str, optional, default=None):
    """

    if not os.path.isfile(input_nifti_file):
        raise FileNotFoundError(f"Input NIfTI file {input_nifti_file} not found.")

    if copy_pose_file is not None:
        if not os.path.isfile(copy_pose_file):
            raise FileNotFoundError(f"Copy pose file {input_nifti_file} not found.")

    made_output_dir, output_dir = ensure_output_file_dir_existence(output_file)

    tmp_dicom_dir = os.path.join(output_dir, f"{uuid1()}")
    try:
        r1 = nifti2dicom(input_nifti_file, tmp_dicom_dir, modality=modality, orientation="ARI", force_overwrite=True)
        r2 = dicom2cavass(tmp_dicom_dir, output_file, offset, copy_pose_file)
    except Exception as e:
        if made_output_dir and os.path.isdir(output_dir):
            shutil.rmtree(output_dir)
        if os.path.isdir(tmp_dicom_dir):
            shutil.rmtree(tmp_dicom_dir)
        if os.path.exists(output_file):
            os.remove(output_file)

        raise e
    shutil.rmtree(tmp_dicom_dir)
    return r1, r2


def cavass2nifti(input_file, output_file, orientation="ARI"):
    """
    Convert cavass IM0 and BIM formats to NIfTI.

    Args:
        input_file (str):
        output_file (str):
        orientation (str, optional, default="ARI"): image orientation of NIfTI file, "ARI" or "LPI"

    Returns:

    """

    if not os.path.isfile(input_file):
        raise FileNotFoundError(f"Input file {input_file} not found.")

    spacing = get_voxel_spacing(input_file)
    data = read_cavass_file(input_file)
    save_nifti(output_file, data, spacing, orientation=orientation)


def nifti_label2cavass(input_nifti_file, output_file, objects,
                       modality=Modality.CT, discard_background=True, copy_pose_file=None):
    """
    Convert NIfTI format segmentation file to cavass BIM format file. A NIfTI file in where contains arbitrary categories
    of objects will convert to multiple CAVASS BIM files, which matches to the number of object categories.

    Args:
        input_nifti_file (str):
        output_file (str): the final saved file for category i in input segmentation will be
        `output_file_prefix_{objects[i]}.BIM`
        objects (sequence or str): objects is an array or a string with comma splitter of object categories,
        where the index of the category in the array is the number that indicates the category in the segmentation.
        modality (Modality, optional, default=Modality.CT):
        discard_background (bool, optional, default True): if True, the regions with label of 0 in the segmentation
        (typically refer to the background region) will not be saved.
        copy_pose_file (str, optional, default=None):

    Returns:

    """
    import nibabel as nib

    if not os.path.isfile(input_nifti_file):
        raise FileNotFoundError(f"Input NIfTI file {input_nifti_file} not found.")

    if copy_pose_file is not None:
        if not os.path.isfile(copy_pose_file):
            raise FileNotFoundError(f"Copy pose file {copy_pose_file} not found.")

    input_data = nib.load(input_nifti_file)
    image_data = input_data.get_fdata()

    if isinstance(objects, str):
        objects = objects.split(",")
    n_classes = len(objects) + 1 if discard_background else len(objects)
    one_hot_arr = one_hot(image_data, num_classes=n_classes)

    start = 1 if discard_background else 0
    for i in range(start, one_hot_arr.shape[3]):
        nifti_label_image = nib.Nifti1Image(one_hot_arr[..., i], input_data.affine, input_data.header, dtype=np.uint8)
        if discard_background:
            obj = objects[i - 1]
        else:
            obj = objects[i]
        tmp_nifti_file = f"{output_file}_{obj}.nii.gz"
        made_output_dir, output_dir = ensure_output_file_dir_existence(tmp_nifti_file)
        try:
            nib.save(nifti_label_image, tmp_nifti_file)
            nifti2cavass(tmp_nifti_file, f"{output_file}_{obj}.BIM", modality, copy_pose_file=copy_pose_file)
        except Exception as e:
            if made_output_dir and os.path.isdir(output_dir):
                shutil.rmtree(output_dir)
            if os.path.exists(tmp_nifti_file):
                os.remove(tmp_nifti_file)
            raise e
        os.remove(tmp_nifti_file)



import os

import numpy as np


def read_mat(input_file, key="scene"):
    from scipy.io import loadmat
    if not os.path.isfile(input_file):
        raise FileNotFoundError(f"Input file {input_file} not found.")
    data = loadmat(input_file)[key]
    return data


def save_mat(output_file, data, key="scene"):
    from scipy.io import savemat
    ensure_output_file_dir_existence(output_file)
    savemat(output_file, {key: data})


def save_nifti(output_file,
               data,
               voxel_spacing=None,
               orientation="LPI"):
    """
    Save image with nii format.

    Args:
        output_file (str):
        data (numpy.ndarray):
        voxel_spacing (sequence or None, optional, default=None): `tuple(x, y, z)`. Voxel spacing of each axis. If None,
            make `voxel_spacing` as `(1.0, 1.0, 1.0)`.
        orientation (str, optional, default="LPI"): "LPI" | "ARI". LPI: Left-Posterior-Inferior;
            ARI: Anterior-Right-Inferior.

    Returns:

    """
    if voxel_spacing is None:
        voxel_spacing = (1.0, 1.0, 1.0)  # replace this with your desired voxel spacing in millimeters

    match orientation:
        case "LPI":
            affine_matrix = np.diag(list(voxel_spacing) + [1.0])
        case "ARI":
            # calculate the affine matrix based on the desired voxel spacing and ARI orientation
            affine_matrix = np.array([
                [0, -voxel_spacing[0], 0, 0],
                [-voxel_spacing[1], 0, 0, 0],
                [0, 0, voxel_spacing[2], 0],
                [0, 0, 0, 1]
            ])
        case _:
            raise ValueError(f"Unsupported orientation {orientation}.")

    # create a NIfTI image object
    import nibabel as nib
    ensure_output_file_dir_existence(output_file)
    nii_img = nib.Nifti1Image(data, affine=affine_matrix)
    nib.save(nii_img, output_file)


def ensure_output_dir_existence(output_dir):
    mk_output_dir = not os.path.exists(output_dir)
    if mk_output_dir:
        os.makedirs(output_dir)
    return mk_output_dir, output_dir


def ensure_output_file_dir_existence(output_file):
    output_dir = os.path.split(output_file)[0]
    return ensure_output_dir_existence(output_dir)



import numpy as np


def windowing(input_data, center, window_width, invert=False):
    min_ = (2 * center - window_width) / 2 + 0.5
    max_ = (2 * center + window_width) / 2 + 0.5
    factor = 255 / (max_ - min_)

    data = (input_data - min_) * factor

    data = np.where(data < 0, 0, data)
    data = np.where(data > 255, 255, data)
    data = data.astype(np.uint8)

    if invert:
        data = 255 - data

    return data


def cavass_soft_tissue_windowing(input_data):
    return windowing(input_data, 1000, 500)


def cavass_bone_windowing(input_data):
    return windowing(input_data, 2000, 4000)


def cavass_pet_windowing(input_data):
    return windowing(input_data, 1200, 3500, invert=True)
